@@ -1,6 +1,6 @@
-/* rhash.c - implementation of rhash library calls
+/* rhash.c - implementation of LibRHash library calls
  *
- * Copyright: 2008 Alexey Kravchenko <rhash.admin@gmail.com>
+ * Copyright: 2008 Aleksey Kravchenko <rhash.admin@gmail.com>
  *
  * Permission is hereby granted,  free of charge,  to any person  obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -16,7 +16,7 @@
 #include <assert.h>
 #include <errno.h>
 
-/* modifier for Windows dll */
+/* modifier for Windows DLL */
 #if defined(_WIN32) && defined(RHASH_EXPORTS)
 # define RHASH_API __declspec(dllexport)
 #endif
@@ -29,6 +29,13 @@
 #include "hex.h"
 #include "rhash.h" /* RHash library interface */
 
+#define STATE_ACTIVE  0xb01dbabe
+#define STATE_STOPED  0xdeadbeef
+#define STATE_DELETED 0xdecea5ed
+#define RCTX_AUTO_FINAL 0x1
+#define RCTX_FINALIZED  0x2
+#define RCTX_FINALIZED_MASK (RCTX_AUTO_FINAL | RCTX_FINALIZED)
+
 /**
  * Initialize static data of rhash algorithms
  */
@@ -40,16 +47,6 @@ void rhash_library_init(void)
 #endif
 }
 
-#if 0
-/**
- * Free allocated resources
- */
-void rhash_library_free(void)
-{
-	rhash_free_algorithms();
-}
-#endif
-
 /**
  * Returns the number of supported hash algorithms.
  *
@@ -60,12 +57,35 @@ int RHASH_API rhash_count(void)
 	return rhash_info_size;
 }
 
-/* basic (lo-level) rhash library functions */
+/**
+ * Information on a hash function and its context
+ */
+typedef struct rhash_vector_item
+{
+	struct rhash_hash_info* hash_info;
+	void *context;
+} rhash_vector_item;
 
 /**
- * Allocate and initialize RHASH library context.
- * After Initializing one should call rhash_update/rhash_final functions.
- * The context must be later freed by calling rhash_free.
+ * The rhash context containing contexts for several hash functions
+ */
+typedef struct rhash_context_ext
+{
+	struct rhash_context rc;
+	unsigned hash_vector_size; /* number of contained hash sums */
+	unsigned flags;
+	unsigned state;
+	void *callback, *callback_data;
+	void *bt_ctx;
+	rhash_vector_item vector[1]; /* contexts of contained hash sums */
+} rhash_context_ext;
+
+/* Lo-level rhash library functions */
+
+/**
+ * Allocate and initialize RHash context for calculating hash(es).
+ * After initializing rhash_update()/rhash_final() functions should be used.
+ * Then the context must be freed by calling rhash_free().
  *
  * @param hash_id union of bit flags, containing ids of hashes to calculate.
  * @return initialized rhash context
@@ -74,8 +94,8 @@ RHASH_API rhash rhash_init(unsigned hash_id)
 {
 	unsigned tail_bit_index; /* index of hash_id trailing bit */
 	unsigned num = 0;        /* number of hashes to compute */
-	rhash_context *rctx = NULL; /* allocated rhash context */
-	size_t hash_size_sum = 0;   /* size of hash contexes to store in rctx */
+	rhash_context_ext *rctx = NULL; /* allocated rhash context */
+	size_t hash_size_sum = 0;   /* size of hash contexts to store in rctx */
 
 	unsigned i, bit_index, id;
 	struct rhash_hash_info* info;
@@ -112,24 +132,26 @@ RHASH_API rhash rhash_init(unsigned hash_id)
 	}
 
 	/* align the size of the rhash context common part */
-	aligned_size = (offsetof(rhash_context, vector[num]) + 7) & ~7;
-	assert(aligned_size >= sizeof(rhash_context));
+	aligned_size = (offsetof(rhash_context_ext, vector[num]) + 7) & ~7;
+	assert(aligned_size >= sizeof(rhash_context_ext));
 
-	/* allocate rhash context with enough memory to store contexes of all used hashes */
-	rctx = (rhash_context*)rsh_malloc(aligned_size + hash_size_sum);
+	/* allocate rhash context with enough memory to store contexts of all used hashes */
+	rctx = (rhash_context_ext*)malloc(aligned_size + hash_size_sum);
+	if(rctx == NULL) return NULL;
 
 	/* initialize common fields of the rhash context */
-	rctx->msg_size = 0;
-	rctx->hash_id = hash_id;
+	memset(rctx, 0, sizeof(rhash_context_ext));
+	rctx->rc.hash_id = hash_id;
+	rctx->flags = RCTX_AUTO_FINAL; /* turn on auto-final by default */
+	rctx->state = STATE_ACTIVE;
 	rctx->hash_vector_size = num;
-	rctx->callback = rctx->callback_data = NULL;
 
-	/* alligned hash contextes follows rctx->vector[num] in the same memory block */
+	/* aligned hash contexts follows rctx->vector[num] in the same memory block */
 	phash_ctx = (char*)rctx + aligned_size;
 	assert(phash_ctx >= (char*)&rctx->vector[num]);
 
 	/* initialize context for every hash in a loop */
-	for(bit_index = tail_bit_index, id = 1 << tail_bit_index, i = 0; 
+	for(bit_index = tail_bit_index, id = 1 << tail_bit_index, i = 0;
 		id <= hash_id; bit_index++, id = id << 1)
 	{
 		/* check if a hash function with given id shall be included into rctx */
@@ -152,74 +174,84 @@ RHASH_API rhash rhash_init(unsigned hash_id)
 		}
 	}
 
-	return rctx; /* return allocated and initialized rhash context */
+	return &rctx->rc; /* return allocated and initialized rhash context */
 }
 
 /**
- * Free RHash contex memory.
+ * Free RHash context memory.
  *
  * @param ctx the context to free.
  */
 void rhash_free(rhash ctx)
 {
+	rhash_context_ext* const ectx = (rhash_context_ext*)ctx;
 	unsigned i;
-	if(!ctx) return;
-	assert(ctx->hash_vector_size <= RHASH_HASH_COUNT);
+	
+	if(ctx == 0) return;
+	assert(ectx->hash_vector_size <= RHASH_HASH_COUNT);
+	ectx->state = STATE_DELETED; /* mark memory block as being removed */
 
 	/* clean the hash functions, which require additional clean up */
-	for(i = 0; i < ctx->hash_vector_size; i++) {
-		struct rhash_hash_info* info = ctx->vector[i].hash_info;
+	for(i = 0; i < ectx->hash_vector_size; i++) {
+		struct rhash_hash_info* info = ectx->vector[i].hash_info;
 		if(info->cleanup != 0) {
-			info->cleanup(ctx->vector[i].context);
+			info->cleanup(ectx->vector[i].context);
 		}
 	}
 
-	free(ctx);
+	free(ectx);
 }
 
 /**
- * Re-inititialize RHash context to reuse it.
+ * Re-initialize RHash context to reuse it.
  * Useful to speed up processing of many small messages.
  *
  * @param ctx context to reinitialize
  */
 RHASH_API void rhash_reset(rhash ctx)
 {
+	rhash_context_ext* const ectx = (rhash_context_ext*)ctx;
 	unsigned i;
-	assert(ctx->hash_vector_size > 0 && ctx->hash_vector_size < RHASH_HASH_COUNT);
+	assert(ectx->hash_vector_size > 0 && ectx->hash_vector_size < RHASH_HASH_COUNT);
+	ectx->state = STATE_ACTIVE; /* re-activate the structure */
 
 	/* re-initialize every hash in a loop */
-	for(i = 0; i < ctx->hash_vector_size; i++) {
-		struct rhash_hash_info* info = ctx->vector[i].hash_info;
+	for(i = 0; i < ectx->hash_vector_size; i++) {
+		struct rhash_hash_info* info = ectx->vector[i].hash_info;
 		if(info->cleanup != 0) {
-			info->cleanup(ctx->vector[i].context);
+			info->cleanup(ectx->vector[i].context);
 		}
 
 		assert(info->init != NULL);
-		info->init(ctx->vector[i].context);
+		info->init(ectx->vector[i].context);
 	}
+	ectx->flags &= ~RCTX_FINALIZED; /* clear finalized state */
 }
 
 /**
  * Calculate hashes of message.
  * Can be called repeatedly with chunks of the message to be hashed.
  *
- * @param pointer to the rhash context
- * @param msg message chunk
- * @param size length of the message chunk
+ * @param ctx the rhash context
+ * @param message message chunk
+ * @param length length of the message chunk
  * @return 0 on success; On fail return -1 and set errno
  */
 RHASH_API int rhash_update(rhash ctx, const void* message, size_t length)
 {
+	rhash_context_ext* const ectx = (rhash_context_ext*)ctx;
 	unsigned i;
-	assert(ctx->hash_vector_size <= RHASH_HASH_COUNT);
+	
+	assert(ectx->hash_vector_size <= RHASH_HASH_COUNT);
+	if(ectx->state != STATE_ACTIVE) return 0; /* do nothing if canceled */
+
 	ctx->msg_size += length;
 
 	/* call update method for every algorithm */
-	for(i = 0; i < ctx->hash_vector_size; i++) {
-		struct rhash_hash_info* info = ctx->vector[i].hash_info;
+	for(i = 0; i < ectx->hash_vector_size; i++) {
+		struct rhash_hash_info* info = ectx->vector[i].hash_info;
 		assert(info->update != 0);
-		info->update(ctx->vector[i].context, message, length);
+		info->update(ectx->vector[i].context, message, length);
 	}
 	return 0; /* no error processing at the moment */
 }
@@ -227,7 +259,7 @@ RHASH_API int rhash_update(rhash ctx, const void* message, size_t length)
 /**
  * Finalize hash calculation and optionally store the first hash.
  *
- * @param ctx rhash context
+ * @param ctx the rhash context
  * @param first_result optional buffer to store a calculated hash with the lowest available id
  * @return 0 on success; On fail return -1 and set errno
  */
@@ -236,47 +268,59 @@ RHASH_API int rhash_final(rhash ctx, unsigned char* first_result)
 	unsigned i = 0;
 	unsigned char buffer[130];
 	unsigned char* out = (first_result ? first_result : buffer);
-	assert(ctx->hash_vector_size <= RHASH_HASH_COUNT);
+	rhash_context_ext* const ectx = (rhash_context_ext*)ctx;
+	assert(ectx->hash_vector_size <= RHASH_HASH_COUNT);
+
+	/* skip final call if already finalized and auto-final is on */
+	if((ectx->flags & RCTX_FINALIZED_MASK) ==
+		(RCTX_AUTO_FINAL | RCTX_FINALIZED)) return 0;
 
 	/* call final method for every algorithm */
-	for(i = 0; i < ctx->hash_vector_size; i++) {
-		struct rhash_hash_info* info = ctx->vector[i].hash_info;
+	for(i = 0; i < ectx->hash_vector_size; i++) {
+		struct rhash_hash_info* info = ectx->vector[i].hash_info;
 		assert(info->final != 0);
 		assert(info->info.digest_size < sizeof(buffer));
-		info->final(ctx->vector[i].context, out);
+		info->final(ectx->vector[i].context, out);
 		out = buffer;
 	}
+	ectx->flags |= RCTX_FINALIZED;
 	return 0; /* no error processing at the moment */
 }
 
 /**
- * Store digest for given hash_id. 
+ * Store digest for given hash_id.
  * If hash_id is zero, function stores digest for a hash with the lowest id found in the context.
- * For nonzero hash_id the context must contain it, otherwise function siliently does nothing.
+ * For nonzero hash_id the context must contain it, otherwise function silently does nothing.
  *
  * @param ctx rhash context
- * @param hash_id id of hash to retrive or zero for hash with the lowest available id
+ * @param hash_id id of hash to retrieve or zero for hash with the lowest available id
  * @param result buffer to put the hash into
  */
 static void rhash_put_digest(rhash ctx, unsigned hash_id, unsigned char* result)
 {
+	rhash_context_ext* const ectx = (rhash_context_ext*)ctx;
 	unsigned i;
 	rhash_vector_item *item;
 	struct rhash_hash_info* info;
 	unsigned char* digest;
 
-	assert(ctx);
-	assert(ctx->hash_vector_size > 0 && ctx->hash_vector_size <= RHASH_HASH_COUNT);
+	assert(ectx);
+	assert(ectx->hash_vector_size > 0 && ectx->hash_vector_size <= RHASH_HASH_COUNT);
+
+	/* finalize context if not yet finalized and auto-final is on */
+	if((ectx->flags & RCTX_FINALIZED_MASK) == RCTX_AUTO_FINAL) {
+		rhash_final(ctx, NULL);
+	}
 
 	if(hash_id == 0) {
-		item = &ctx->vector[0]; /* get the first hash */
+		item = &ectx->vector[0]; /* get the first hash */
 		info = item->hash_info;
 	} else {
 		for(i = 0;; i++) {
-			if(i >= ctx->hash_vector_size) {
+			if(i >= ectx->hash_vector_size) {
 				return; /* hash_id not found, do nothing */
 			}
-			item = &ctx->vector[i];
+			item = &ectx->vector[i];
 			info = item->hash_info;
 			if(info->info.hash_id == hash_id) break;
 		}
@@ -292,10 +336,10 @@ static void rhash_put_digest(rhash ctx, unsigned hash_id, unsigned char* result)
 }
 
 /**
- * Set the callback function to be called from the 
+ * Set the callback function to be called from the
  * rhash_file() and rhash_file_update() functions
  * on processing every file block. The file block
- * size is set internaly by rhash and now is 8 KiB.
+ * size is set internally by rhash and now is 8 KiB.
  *
  * @param ctx rhash context
  * @param callback pointer to the callback function
@@ -303,8 +347,8 @@ static void rhash_put_digest(rhash ctx, unsigned hash_id, unsigned char* result)
  */
 RHASH_API void rhash_set_callback(rhash ctx, rhash_callback_t callback, void* callback_data)
 {
-	ctx->callback = callback;
-	ctx->callback_data = callback_data;
+	((rhash_context_ext*)ctx)->callback = callback;
+	((rhash_context_ext*)ctx)->callback_data = callback_data;
 }
 
 
@@ -316,7 +360,7 @@ RHASH_API void rhash_set_callback(rhash ctx, rhash_callback_t callback, void* ca
  * @param hash_id id of hash sum to compute
  * @param message the message to process
  * @param length message length
- * @param result buffer to recieve bynary hash string
+ * @param result buffer to receive binary hash string
  * @return 0 on success, -1 on error
  */
 RHASH_API int rhash_msg(unsigned hash_id, const void* message, size_t length, unsigned char* result)
@@ -334,7 +378,7 @@ RHASH_API int rhash_msg(unsigned hash_id, const void* message, size_t length, un
 /**
  * Hash a file or stream. Multiple hashes can be computed.
  * First, inintialize ctx parameter with rhash_init() before calling
- * rhash_file_update(). Then use rhash_final() and rhash_print() 
+ * rhash_file_update(). Then use rhash_final() and rhash_print()
  * to retrive hash values. Finaly call rhash_free() on ctx
  * to free allocated memory or call rhash_reset() to reuse ctx.
  *
@@ -344,21 +388,27 @@ RHASH_API int rhash_msg(unsigned hash_id, const void* message, size_t length, un
  */
 RHASH_API int rhash_file_update(rhash ctx, FILE* fd)
 {
+	rhash_context_ext* const ectx = (rhash_context_ext*)ctx;
 	const size_t block_size = 8192;
 	unsigned char *buffer, *pmem;
 	size_t length = 0, align8;
 	int res = 0;
+	if(ectx->state != STATE_ACTIVE) return 0; /* do nothing if canceled */
 
 	if(ctx == NULL) {
 		errno = EINVAL;
 		return -1;
 	}
 
-	pmem = (unsigned char*)rsh_malloc(block_size + 8);
+	pmem = (unsigned char*)malloc(block_size + 8);
+	if(!pmem) return -1; /* errno is set to ENOMEM according to UNIX 98 */
+
 	align8 = ((unsigned char*)0 - pmem) & 7;
 	buffer = pmem + align8;
 
 	while(!feof(fd)) {
+		if(ectx->state != STATE_ACTIVE) break; /* stop if canceled */
+
 		length = fread(buffer, 1, block_size, fd);
 		/* read can return -1 on error */
 		if(length == (size_t)-1) {
@@ -367,8 +417,9 @@ RHASH_API int rhash_file_update(rhash ctx, FILE* fd)
 		}
 		rhash_update(ctx, buffer, length);
 
-		if(ctx->callback)
-			((rhash_callback_t)ctx->callback)(ctx->callback_data, ctx->msg_size);
+		if(ectx->callback) {
+			((rhash_callback_t)ectx->callback)(ectx->callback_data, ectx->rc.msg_size);
+		}
 	}
 
 	free(buffer);
@@ -462,7 +513,7 @@ rhash_info* rhash_info_by_id(unsigned hash_id)
  */
 RHASH_API int rhash_is_base32(unsigned hash_id)
 {
-	/* fast method is just to test a bitmask */
+	/* fast method is just to test a bit-mask */
 	return ((hash_id & (RHASH_TTH | RHASH_AICH)) != 0);
 }
 
@@ -488,7 +539,7 @@ RHASH_API int rhash_get_digest_size(unsigned hash_id)
 RHASH_API int rhash_get_hash_length(unsigned hash_id)
 {
 	rhash_info* info = rhash_info_by_id(hash_id);
-	return (int)(info ? (info->flags & F_BS32 ? 
+	return (int)(info ? (info->flags & F_BS32 ?
 		BASE32_LENGTH(info->digest_size) : info->digest_size * 2) : 0);
 }
 
@@ -507,20 +558,22 @@ RHASH_API const char* rhash_get_name(unsigned hash_id)
 /* hash sum output */
 
 /**
- * Print text presentation of a given hash sum to specified buffer
+ * Print a text presentation of a given hash sum to the specified buffer,
  *
  * @param output a buffer to print the hash to
- * @param context algorithms state
- * @param hash_id id of the sum to print
- * @param flags  controls how to print the sum, can contain flags 
- *               RHPR_UPPERCASE, RHPR_HEX, RHPR_BASE32, e.t.c.
- * @return number of writen characters on success, 0 on fail
+ * @param bytes a hash sum to print
+ * @param size a size of hash sum in bytes
+ * @param flags  a bit-mask controlling how to format the hash sum,
+ *               can be a mix of the flags: RHPR_RAW, RHPR_HEX, RHPR_BASE32,
+ *               RHPR_BASE64, RHPR_UPPERCASE, RHPR_REVERSE
+ * @return the number of written characters
  */
-size_t rhash_print_bytes(char* output, unsigned char* bytes, size_t size, int flags)
+size_t rhash_print_bytes(char* output, const unsigned char* bytes,
+	size_t size, int flags)
 {
 	size_t str_len;
 	int upper_case = (flags & RHPR_UPPERCASE);
-	int format = (flags & ~(RHPR_UPPERCASE|RHPR_REVERSE));
+	int format = (flags & ~(RHPR_UPPERCASE | RHPR_REVERSE));
 
 	switch(format) {
 	case RHPR_HEX:
@@ -552,7 +605,7 @@ size_t rhash_print_bytes(char* output, unsigned char* bytes, size_t size, int fl
  * @param context algorithms state
  * @param hash_id id of the hash sum to print or 0 to print the first hash
  *                saved in the context.
- * @param flags  controls how to print the sum, can contain flags 
+ * @param flags  controls how to print the sum, can contain flags
  *               RHPR_UPPERCASE, RHPR_HEX, RHPR_BASE32, RHPR_BASE64, e.t.c.
  * @return number of writen characters on success, 0 on fail
  */
@@ -563,7 +616,7 @@ size_t RHASH_API rhash_print(char* output, rhash context, unsigned hash_id, int
 	size_t digest_size;
 
 	info = (hash_id != 0 ? rhash_info_by_id(hash_id) :
-		&context->vector[0].hash_info->info);
+		&((rhash_context_ext*)context)->vector[0].hash_info->info);
 
 	if(info == NULL) return 0;
 	digest_size = info->digest_size;
@@ -613,10 +666,10 @@ BOOL APIENTRY DllMain(HMODULE hModule, DWORD reason, LPVOID reserved)
 #endif
 
 /**
- * Process a bittorent-related rhash message.
+ * Process a BitTorrent-related rhash message.
  *
  * @param msg_id message identifier
- * @param bt bittorent context
+ * @param bt BitTorrent context
  * @param ldata data depending on message
  * @param rdata data depending on message
  * @return message-specific data
@@ -649,6 +702,8 @@ static rhash_uptr_t process_bt_msg(unsigned msg_id, torrent_ctx* bt, rhash_uptr_
 	return 0;
 }
 
+#define PVOID2UPTR(p) ((rhash_uptr_t)((char*)p - 0))
+
 /**
  * Process a rhash message.
  *
@@ -660,12 +715,12 @@ static rhash_uptr_t process_bt_msg(unsigned msg_id, torrent_ctx* bt, rhash_uptr_
  */
 RHASH_API rhash_uptr_t rhash_transmit(unsigned msg_id, void* dst, rhash_uptr_t ldata, rhash_uptr_t rdata)
 {
-#define PVOID2UPTR(p) ((rhash_uptr_t)((char*)p - 0))
+	/* for messages working with rhash context */
+	rhash_context_ext* const ctx = (rhash_context_ext*)dst;
 
 	switch(msg_id) {
 	case RMSG_GET_CONTEXT:
 		{
-			rhash ctx = (rhash)dst;
 			unsigned i;
 			for(i = 0; i < ctx->hash_vector_size; i++) {
 				struct rhash_hash_info* info = ctx->vector[i].hash_info;
@@ -675,7 +730,22 @@ RHASH_API rhash_uptr_t rhash_transmit(unsigned msg_id, void* dst, rhash_uptr_t l
 			return (rhash_uptr_t)0;
 		}
 
-		/* openssl related messages */
+	case RMSG_CANCEL:
+		/* mark rhash context as canceled, in a multithreaded program */
+		atomic_compare_and_swap(&ctx->state, STATE_ACTIVE, STATE_STOPED);
+		return 0;
+
+	case RMSG_IS_CANCELED:
+		return (ctx->state == STATE_STOPED);
+
+	case RMSG_GET_FINALIZED:
+		return ((ctx->flags & RCTX_FINALIZED) != 0);
+	case RMSG_SET_AUTOFINAL:
+		ctx->flags &= ~RCTX_AUTO_FINAL;
+		if(ldata) ctx->flags |= RCTX_AUTO_FINAL;
+		break;
+
+	/* OpenSSL related messages */
 #ifdef USE_OPENSSL
 	case RMSG_SET_OPENSSL_MASK:
 		rhash_openssl_hash_mask = (unsigned)ldata;
@@ -684,14 +754,14 @@ RHASH_API rhash_uptr_t rhash_transmit(unsigned msg_id, void* dst, rhash_uptr_t l
 		return rhash_openssl_hash_mask;
 #endif
 
-		/* BitTorrent related messages */
+	/* BitTorrent related messages */
 	case RMSG_BT_ADD_FILE:
 	case RMSG_BT_SET_OPTIONS:
 	case RMSG_BT_SET_ANNOUNCE:
 	case RMSG_BT_SET_PIECE_LENGTH:
 	case RMSG_BT_SET_PROGRAM_NAME:
 	case RMSG_BT_GET_TEXT:
-		return process_bt_msg(msg_id, (torrent_ctx*)(((rhash_context*)dst)->bt_ctx), ldata, rdata);
+		return process_bt_msg(msg_id, (torrent_ctx*)(((rhash_context_ext*)dst)->bt_ctx), ldata, rdata);
 
 	default:
 		return RHASH_ERROR; /* unknown message */
