@@ -12,10 +12,9 @@
 #endif
 
 #include "librhash/rhash.h"
-#include "librhash/plug_openssl.h"
 #include "win_utils.h"
 #include "file_mask.h"
-#include "crc_print.h"
+#include "hash_print.h"
 #include "output.h"
 #include "rhash_main.h"
 #include "version.h"
@@ -33,11 +32,15 @@ struct options_t opt;      /* command line options */
 static void print_help(void)
 {
 	assert(rhash_data.out != NULL);
-	fprintf(rhash_data.out, "%s\n%s", VERSION_STRING,
-		"Usage: " CMD_FILENAME " [<option> ...] <filename|-> [...]\n"
-		"       " CMD_FILENAME " --printf=<format string> <filename|-> [...]\n\n"
-		"Options:\n"
-		"  -V, --version Print " PROGRAM_NAME " version and exit.\n"
+
+	/* print program version and usage */
+	fprintf(rhash_data.out, _("%s\n"
+		"Usage: %s [OPTION...] [FILE | -]...\n"
+		"       %s --printf=<format string> [FILE | -]...\n\n"), VERSION_STRING, CMD_FILENAME, CMD_FILENAME);
+
+	fprintf(rhash_data.out, "%s%s%s",
+		_("Options:\n"
+		"  -V, --version Print program version and exit.\n"
 		"  -h, --help    Print this help screen.\n"
 		"  -C, --crc32   Calculate CRC32 hash sum.\n"
 		"      --md4     Calculate MD4   hash sum.\n"
@@ -76,13 +79,15 @@ static void print_help(void)
 		"      --bsd     Print hash sums, using BSD-like format.\n"
 		"      --simple  Print hash sums, using simple format.\n"
 		"  -m, --magnet  Print hash sums  as magnet links.\n"
-		"      --torrent Create torrent files.\n"
+		"      --torrent Create torrent files.\n"),
 #ifdef _WIN32
-		"      --ansi    Use Windows codepage for output (Windows only).\n"
+		_("      --ansi    Use Windows codepage for output (Windows only).\n"),
+#else
+		"",
 #endif
-		"      --template=<file> Load a printf-like template from the <file>\n"
+		_("      --template=<file> Load a printf-like template from the <file>\n"
 		"  -p, --printf=<format string>  Format and print hash sums.\n"
-		"                See the RHash manual for details.\n"
+		"                See the RHash manual for details.\n")
 		);
 	rsh_exit(0);
 }
@@ -93,7 +98,6 @@ static void print_help(void)
 static void list_hashes(void)
 {
 	int id;
-	assert(rhash_data.out != NULL);
 	for(id = 1; id < RHASH_ALL_HASHES; id <<= 1) {
 		const char* hash_name = rhash_get_name(id);
 		if(hash_name) fprintf(rhash_data.out, "%s\n", hash_name);
@@ -103,6 +107,10 @@ static void list_hashes(void)
 
 /**
  * Process --accept and --crc-accept options.
+ *
+ * @param o pointer to the options structure to update
+ * @param accept_string comma delimited string to parse
+ * @param type non-zero for the crc_accept option
  */
 static void crc_accept(options_t *o, char* accept_string, unsigned type)
 {
@@ -113,13 +121,17 @@ static void crc_accept(options_t *o, char* accept_string, unsigned type)
 
 /**
  * Process an --openssl option.
+ *
+ * @param o pointer to the options structure to update
+ * @param openssl_hashes comma delimited string with hash names
+ * @param type ignored
  */
 static void openssl_flags(options_t *o, char* openssl_hashes, unsigned type)
 {
 #ifdef USE_OPENSSL
 	char *cur, *next;
 	(void)type;
-	o->openssl_mask = 0x80000000;
+	o->openssl_mask = 0x80000000; /* turn off using default mask */
 
 	/* set the openssl_mask */
 	for(cur = openssl_hashes; cur && *cur; cur = next) {
@@ -127,10 +139,10 @@ static void openssl_flags(options_t *o, char* openssl_hashes, unsigned type)
 		unsigned bit;
 		size_t length;
 		next = strchr(cur, ',');
-		length = (next != NULL ? (size_t)(next - cur) : strlen(cur));
+		length = (next != NULL ? (size_t)(next++ - cur) : strlen(cur));
 
 		for(bit = 1; bit <= RHASH_ALL_HASHES; bit = bit << 1, info++) {
-			if( (bit && OPENSSL_SUPPORTED_HASHES_MASK) &&
+			if( (bit & RHASH_OPENSSL_SUPPORTED_HASHES) &&
 				memcmp(cur, info->short_name, length) == 0 &&
 				info->short_name[length] == 0) {
 					o->openssl_mask |= bit;
@@ -138,20 +150,22 @@ static void openssl_flags(options_t *o, char* openssl_hashes, unsigned type)
 			}
 		}
 		if(bit > RHASH_ALL_HASHES) {
-			cur[length] = 0;
-			log_msg(PROGRAM_NAME " warning: openssl option doesn't support '%s' hash\n", cur);
+			cur[length] = '\0'; /* terminate wrong hash name */
+			log_warning(_("openssl option doesn't support '%s' hash\n"), cur);
 		}
 	}
 #else
 	(void)type;
 	(void)openssl_hashes;
 	(void)o;
-	log_msg(PROGRAM_NAME " warning: compiled without openssl support\n");
+	log_warning(_("compiled without openssl support\n"));
 #endif
 }
 
 /**
  * Process --video option.
+ *
+ * @param o pointer to the options structure to update
  */
 static void accept_video(options_t *o)
 {
@@ -163,9 +177,8 @@ static void accept_video(options_t *o)
  */
 static void nya(void)
 {
-	assert(rhash_data.out != NULL);
-	fprintf(rhash_data.out, "\n  /\\___/\\   _\n = ' T ' = //\n"
-		"  |     |\\//\n  | | | | )\n  \"\"  \"\" \"\n");
+	fprintf(rhash_data.out, "  /\\__/\\\n (^ _ ^.) %s\n  (_uu__)\n",
+		_("Purrr..."));
 	rsh_exit(0);
 }
 
@@ -173,31 +186,31 @@ static void nya(void)
  * Process on --maxdepth option.
  *
  * @param o pointer to the processed option
- * @param number string containing the max-depth number
+ * @param number the string containing the max-depth number
  * @param param unused parameter
  */
 static void set_max_depth(options_t *o, char* number, unsigned param)
 {
 	(void)param;
 	if(strspn(number, "0123456789") < strlen(number)) {
-		log_msg(PROGRAM_NAME ": maxdepth parameter is not a number: %s\n", number);
+		log_error(_("maxdepth parameter is not a number: %s\n"), number);
 		rsh_exit(2);
 	}
 	o->find_max_depth = atoi(number);
 }
 
 /**
- * Set bittorent file piece-length
+ * Set the length of a BitTorrent file piece.
  *
  * @param o pointer to the processed option
- * @param number string containing the bt-piece-length number
+ * @param number string containing the piece length number
  * @param param unused parameter
  */
 static void set_bt_piece_length(options_t *o, char* number, unsigned param)
 {
 	(void)param;
 	if(strspn(number, "0123456789") < strlen(number)) {
-		log_msg(PROGRAM_NAME ": bt-piece-length parameter is not a number: %s\n", number);
+		log_error(_("bt-piece-length parameter is not a number: %s\n"), number);
 		rsh_exit(2);
 	}
 	o->bt_piece_length = (size_t)atoi(number);
@@ -216,13 +229,13 @@ static void set_path_separator(options_t *o, char* sep, unsigned param)
 	if((*sep == '/' || *sep == '\\') && sep[1] == 0) {
 		o->path_separator = *sep;
 #if defined(_WIN32)
-		/* MSYS enviropment changes '/' in command line to HOME, see http://www.mingw.org/wiki/FAQ */
+		/* MSYS environment changes '/' in command line to HOME, see http://www.mingw.org/wiki/FAQ */
 	} else if(getenv("MSYSTEM") || getenv("TERM")) {
-		log_msg("warning: wrong path-separator, use '//' instead of '/' on MSYS\n");
+		log_warning(_("wrong path-separator, use '//' instead of '/' on MSYS\n"));
 		o->path_separator = '/';
 #endif
 	} else {
-		log_msg(PROGRAM_NAME ": error: path-separator is not '/' or '\\': %s\n", sep);
+		log_error(_("path-separator is not '/' or '\\': %s\n"), sep);
 		rsh_exit(2);
 	}
 }
@@ -232,109 +245,110 @@ static void set_path_separator(options_t *o, char* sep, unsigned param)
  */
 typedef struct cmdline_opt_t
 {
-	char  type;           /* how to porcess option, see below */
+	unsigned short type;  /* how to process the option, see option_type_t below */
 	char  short1, short2; /* short option names */
 	char* long_name;      /* long option name */
-	void* ptr;            /* accessorial pointer, e.g. to an opt field */
+	void* ptr;            /* auxiliary pointer, e.g. to an opt field */
 	unsigned param;       /* optional integer parameter */
 } cmdline_opt_t;
 
-/* The 'type' character of cmdline_opt_t determines how option will be processed:
- * u - set a bit flag in a uint32_t field of the opt structure
- * s - store the second parameter into a char* field
- * P - store file path (without a codepage conversion)
- * F - call a function to process the following string parameter
- * f - call a function for complex option processing
- * p - print a constant string and exit
- * Note, that only option of types 's', 'P', 'F' require a second parameter.
- */
+enum option_type_t
+{
+	F_NEED_PARAM = 16, /* flag: option needs a parameter */
+	F_OUTPUT_OPT = 32, /* flag: option changes program output */
+	F_UFLG = 1, /* set a bit flag in a uint32_t field */
+	F_UENC = F_UFLG | F_OUTPUT_OPT, /* an encoding changing option */
+	F_CSTR = 2 | F_NEED_PARAM, /* store parameter as a C string */
+	F_OPTH = 3 | F_NEED_PARAM | F_OUTPUT_OPT,
+	F_VFNC = 4, /* just call a function */
+	F_PFNC = 5 | F_NEED_PARAM, /* process option parameter by calling a handler */
+	F_PRNT = 6, /* print a constant C-string and exit */
+};
+
+#define is_param_required(option_type) ((option_type) & F_NEED_PARAM)
+#define is_output_modifier(option_type) ((option_type) & F_OUTPUT_OPT)
 
 /* supported program options */
-cmdline_opt_t cmdline_opt[] = {
+cmdline_opt_t cmdline_opt[] =
+{
 	/* program modes */
-	{ 'u', 'c',   0, "check",  &opt.mode, MODE_CHECK },
-	{ 'u',   0,   0, "check-embedded",  &opt.mode, MODE_CHECK_EMBEDDED },
-	{ 'u', 'u',   0, "update", &opt.mode, MODE_UPDATE },
-	{ 'u', 'B',   0, "benchmark", &opt.mode, MODE_BENCHMARK },
-	{ 'u',   0,   0, "torrent", &opt.mode, MODE_TORRENT },
-	{ 'f',   0,   0, "list-hashes", list_hashes, 0 },
-	{ 'f', 'h',   0, "help",   print_help, 0 },
-	{ 'p', 'V',   0, "version", VERSION_STRING, 0 },
+	{ F_UFLG, 'c',   0, "check",  &opt.mode, MODE_CHECK },
+	{ F_UFLG,   0,   0, "check-embedded",  &opt.mode, MODE_CHECK_EMBEDDED },
+	{ F_UFLG, 'u',   0, "update", &opt.mode, MODE_UPDATE },
+	{ F_UFLG, 'B',   0, "benchmark", &opt.mode, MODE_BENCHMARK },
+	{ F_UFLG,   0,   0, "torrent", &opt.mode, MODE_TORRENT },
+	{ F_VFNC,   0,   0, "list-hashes", list_hashes, 0 },
+	{ F_VFNC, 'h',   0, "help",   print_help, 0 },
+	{ F_PRNT, 'V',   0, "version", VERSION_STRING, 0 },
+
 	/* hash sums options */
-	{ 'u', 'a',   0, "all",    &opt.sum_flags, RHASH_ALL_HASHES },
-	{ 'u', 'C',   0, "crc32",  &opt.sum_flags, RHASH_CRC32 },
-	{ 'u',   0,   0, "md4",    &opt.sum_flags, RHASH_MD4 },
-	{ 'u', 'M',   0, "md5",    &opt.sum_flags, RHASH_MD5 },
-	{ 'u', 'H',   0, "sha1",   &opt.sum_flags, RHASH_SHA1 },
-	{ 'u',   0,   0, "sha224", &opt.sum_flags, RHASH_SHA224 },
-	{ 'u',   0,   0, "sha256", &opt.sum_flags, RHASH_SHA256 },
-	{ 'u',   0,   0, "sha384", &opt.sum_flags, RHASH_SHA384 },
-	{ 'u',   0,   0, "sha512", &opt.sum_flags, RHASH_SHA512 },
-	{ 'u',   0,   0, "tiger",  &opt.sum_flags, RHASH_TIGER },
-	{ 'u', 'T',   0, "tth",    &opt.sum_flags, RHASH_TTH },
-	{ 'u',   0,   0, "btih",   &opt.sum_flags, RHASH_BTIH },
-	{ 'u', 'E',   0, "ed2k",   &opt.sum_flags, RHASH_ED2K },
-	{ 'u', 'A',   0, "aich",   &opt.sum_flags, RHASH_AICH },
-	{ 'u', 'G',   0, "gost",   &opt.sum_flags, RHASH_GOST },
-	{ 'u',   0,   0, "gost-cryptopro", &opt.sum_flags, RHASH_GOST_CRYPTOPRO },
-	{ 'u', 'W',   0, "whirlpool", &opt.sum_flags, RHASH_WHIRLPOOL },
-	{ 'u',   0,   0, "ripemd160", &opt.sum_flags, RHASH_RIPEMD160 },
-	{ 'u',   0,   0, "has160",    &opt.sum_flags, RHASH_HAS160 },
-	{ 'u',   0,   0, "snefru128", &opt.sum_flags, RHASH_SNEFRU128 },
-	{ 'u',   0,   0, "snefru256", &opt.sum_flags, RHASH_SNEFRU256 },
-	{ 'u',   0,   0, "edonr256",  &opt.sum_flags, RHASH_EDONR256 },
-	{ 'u',   0,   0, "edonr512",  &opt.sum_flags, RHASH_EDONR512 },
-	{ 'u', 'L',   0, "ed2k-link", &opt.sum_flags, OPT_ED2K_LINK },
+	{ F_UFLG, 'a',   0, "all",    &opt.sum_flags, RHASH_ALL_HASHES },
+	{ F_UFLG, 'C',   0, "crc32",  &opt.sum_flags, RHASH_CRC32 },
+	{ F_UFLG,   0,   0, "md4",    &opt.sum_flags, RHASH_MD4 },
+	{ F_UFLG, 'M',   0, "md5",    &opt.sum_flags, RHASH_MD5 },
+	{ F_UFLG, 'H',   0, "sha1",   &opt.sum_flags, RHASH_SHA1 },
+	{ F_UFLG,   0,   0, "sha224", &opt.sum_flags, RHASH_SHA224 },
+	{ F_UFLG,   0,   0, "sha256", &opt.sum_flags, RHASH_SHA256 },
+	{ F_UFLG,   0,   0, "sha384", &opt.sum_flags, RHASH_SHA384 },
+	{ F_UFLG,   0,   0, "sha512", &opt.sum_flags, RHASH_SHA512 },
+	{ F_UFLG,   0,   0, "tiger",  &opt.sum_flags, RHASH_TIGER },
+	{ F_UFLG, 'T',   0, "tth",    &opt.sum_flags, RHASH_TTH },
+	{ F_UFLG,   0,   0, "btih",   &opt.sum_flags, RHASH_BTIH },
+	{ F_UFLG, 'E',   0, "ed2k",   &opt.sum_flags, RHASH_ED2K },
+	{ F_UFLG, 'A',   0, "aich",   &opt.sum_flags, RHASH_AICH },
+	{ F_UFLG, 'G',   0, "gost",   &opt.sum_flags, RHASH_GOST },
+	{ F_UFLG,   0,   0, "gost-cryptopro", &opt.sum_flags, RHASH_GOST_CRYPTOPRO },
+	{ F_UFLG, 'W',   0, "whirlpool", &opt.sum_flags, RHASH_WHIRLPOOL },
+	{ F_UFLG,   0,   0, "ripemd160", &opt.sum_flags, RHASH_RIPEMD160 },
+	{ F_UFLG,   0,   0, "has160",    &opt.sum_flags, RHASH_HAS160 },
+	{ F_UFLG,   0,   0, "snefru128", &opt.sum_flags, RHASH_SNEFRU128 },
+	{ F_UFLG,   0,   0, "snefru256", &opt.sum_flags, RHASH_SNEFRU256 },
+	{ F_UFLG,   0,   0, "edonr256",  &opt.sum_flags, RHASH_EDONR256 },
+	{ F_UFLG,   0,   0, "edonr512",  &opt.sum_flags, RHASH_EDONR512 },
+	{ F_UFLG, 'L',   0, "ed2k-link", &opt.sum_flags, OPT_ED2K_LINK },
+
 	/* output formats */
-	{ 'u',   0,   0, "sfv",     &opt.fmt, FMT_SFV },
-	{ 'u',   0,   0, "bsd",     &opt.fmt, FMT_BSD },
-	{ 'u',   0,   0, "simple",  &opt.fmt, FMT_SIMPLE },
-	{ 'u', 'm',   0, "magnet",  &opt.fmt, FMT_MAGNET },
-	{ 'u',   0,   0, "uppercase", &opt.flags, OPT_UPPERCASE },
-	{ 'u',   0,   0, "lowercase", &opt.flags, OPT_LOWERCASE },
-	{ 's',   0,   0, "template",  &opt.template_file, 0 },
-	{ 's', 'p',   0, "printf",  &opt.printf, 0 },
+	{ F_UFLG,   0,   0, "sfv",     &opt.fmt, FMT_SFV },
+	{ F_UFLG,   0,   0, "bsd",     &opt.fmt, FMT_BSD },
+	{ F_UFLG,   0,   0, "simple",  &opt.fmt, FMT_SIMPLE },
+	{ F_UFLG, 'm',   0, "magnet",  &opt.fmt, FMT_MAGNET },
+	{ F_UFLG,   0,   0, "uppercase", &opt.flags, OPT_UPPERCASE },
+	{ F_UFLG,   0,   0, "lowercase", &opt.flags, OPT_LOWERCASE },
+	{ F_CSTR,   0,   0, "template",  &opt.template_file, 0 },
+	{ F_CSTR, 'p',   0, "printf",  &opt.printf_str, 0 },
+
 	/* other options */
-	{ 'u', 'r', 'R', "recursive", &opt.flags, OPT_RECURSIVE },
-	{ 'u', 'v',   0, "verbose", &opt.flags, OPT_VERBOSE },
-	{ 'u',   0,   0, "gost-reverse", &opt.flags, OPT_GOST_REVERSE },
-	{ 'u',   0,   0, "skip-ok", &opt.flags, OPT_SKIP_OK },
-	{ 'u', 'i',   0, "ignore-case", &opt.flags, OPT_IGNORE_CASE },
-	{ 'u',   0,   0, "percents", &opt.flags, OPT_PERCENTS },
-	{ 'u',   0,   0, "speed",  &opt.flags, OPT_SPEED },
-	{ 'u', 'e',   0, "embed-crc",  &opt.flags, OPT_EMBED_CRC },
-	{ 's',   0,   0, "embed-crc-delimiter", &opt.embed_crc_delimiter, 0 },
-	{ 'F',   0,   0, "path-separator", set_path_separator, 0 },
-	{ 'P', 'o',   0, "output", &opt.output, 0 },
-	{ 'P', 'l',   0, "log",    &opt.log,    0 },
-	{ 'F', 'q',   0, "accept", crc_accept, 0 },
-	{ 'F', 't',   0, "crc-accept", crc_accept, 1 },
-	{ 'f',   0,   0, "video",  accept_video, 0 },
-	{ 'f',   0,   0, "nya",  nya, 0 },
-	{ 'F',   0,   0, "maxdepth", set_max_depth, 0 },
-	{ 'u',   0,   0, "bt-private", &opt.flags, OPT_BT_PRIVATE },
-	{ 'F',   0,   0, "bt-piece-length", set_bt_piece_length, 0 },
-	{ 's',   0,   0, "bt-announce", &opt.bt_announce, 0 },
-	{ 'u',   0,   0, "benchmark-raw", &opt.flags, OPT_BENCH_RAW },
-	{ 'F',   0,   0, "openssl", openssl_flags, 0 },
-#ifdef _WIN32 /* code pages */
-	{ 'u',   0,   0, "utf8", &opt.flags, OPT_UTF8 },
-	{ 'u',   0,   0, "ansi", &opt.flags, OPT_ANSI },
-	{ 'u',   0,   0, "oem",  &opt.flags, OPT_OEM },
+	{ F_UFLG, 'r', 'R', "recursive", &opt.flags, OPT_RECURSIVE },
+	{ F_UFLG, 'v',   0, "verbose", &opt.flags, OPT_VERBOSE },
+	{ F_UFLG,   0,   0, "gost-reverse", &opt.flags, OPT_GOST_REVERSE },
+	{ F_UFLG,   0,   0, "skip-ok", &opt.flags, OPT_SKIP_OK },
+	{ F_UFLG, 'i',   0, "ignore-case", &opt.flags, OPT_IGNORE_CASE },
+	{ F_UENC,   0,   0, "percents", &opt.flags, OPT_PERCENTS },
+	{ F_UFLG,   0,   0, "speed",  &opt.flags, OPT_SPEED },
+	{ F_UFLG, 'e',   0, "embed-crc",  &opt.flags, OPT_EMBED_CRC },
+	{ F_CSTR,   0,   0, "embed-crc-delimiter", &opt.embed_crc_delimiter, 0 },
+	{ F_PFNC,   0,   0, "path-separator", set_path_separator, 0 },
+	{ F_OPTH, 'o',   0, "output", &opt.output, 0 },
+	{ F_OPTH, 'l',   0, "log",    &opt.log,    0 },
+	{ F_PFNC, 'q',   0, "accept", crc_accept, 0 },
+	{ F_PFNC, 't',   0, "crc-accept", crc_accept, 1 },
+	{ F_VFNC,   0,   0, "video",  accept_video, 0 },
+	{ F_VFNC,   0,   0, "nya",  nya, 0 },
+	{ F_PFNC,   0,   0, "maxdepth", set_max_depth, 0 },
+	{ F_UFLG,   0,   0, "bt-private", &opt.flags, OPT_BT_PRIVATE },
+	{ F_PFNC,   0,   0, "bt-piece-length", set_bt_piece_length, 0 },
+	{ F_CSTR,   0,   0, "bt-announce", &opt.bt_announce, 0 },
+	{ F_UFLG,   0,   0, "benchmark-raw", &opt.flags, OPT_BENCH_RAW },
+	{ F_PFNC,   0,   0, "openssl", openssl_flags, 0 },
+
+#ifdef _WIN32 /* code pages (windows only) */
+	{ F_UENC,   0,   0, "utf8", &opt.flags, OPT_UTF8 },
+	{ F_UENC,   0,   0, "ansi", &opt.flags, OPT_ANSI },
+	{ F_UENC,   0,   0, "oem",  &opt.flags, OPT_OEM },
 #endif
 	{ 0,0,0,0,0,0 }
 };
 
-/**
- * Return non-zero if given option has a required parameter
- *
- * @param type the type of option to check
- */
-static int is_param_required(char type)
-{
-	return (type == 's' || type == 'P' || type == 'F');
-}
-
 /**
  * Log a message and exit the program.
  *
@@ -342,7 +356,7 @@ static int is_param_required(char type)
  */
 static void die(const char* msg)
 {
-	log_msg(msg);
+	log_error(msg);
 	rsh_exit(2);
 }
 
@@ -353,7 +367,7 @@ static void die(const char* msg)
  */
 static void fail_on_unknow_option(const char* option_name)
 {
-	log_msg(PROGRAM_NAME ": unknown option: %s", (option_name ? option_name : "?"));
+	log_error(_("unknown option: %s"), (option_name ? option_name : "?"));
 	rsh_exit(2);
 }
 
@@ -369,55 +383,63 @@ typedef struct parsed_option_t
 /**
  * Process given command line option
  *
- * @param opts the sructure to store results of option processing
+ * @param opts the structure to store results of option processing
  * @param option option to process
  */
 static void apply_option(options_t *opts, parsed_option_t* option)
 {
 	cmdline_opt_t* o = option->o;
-	char type = o->type;
-	char* value;
-
-	if(type == 'u') {
-		*(unsigned*)((char*)opts + ((char*)o->ptr - (char*)&opt)) |= o->param;
-	}
-	else if(type == 's' || type == 'F' || type == 'P')
-	{
-		/* option requires a parameter */
+	char option_type = o->type;
+	char* value = NULL;
+	
+	/* check if option requires a parameter */
+	if(is_param_required(option_type)) {
 		if(!option->parameter) {
-			log_msg(PROGRAM_NAME ": argument is required for option %s\n", option->name);
+			log_error(_("argument is required for option %s\n"), option->name);
 			rsh_exit(2);
 		}
-
 #ifdef _WIN32
 		/* convert from UTF-16 if not a filepath */
-		if(type != 'P') {
+		if(option_type != F_OPTH) {
 			value = w2c((wchar_t*)option->parameter);
 			rsh_vector_add_ptr(opt.mem, value);
-		} else 
+		} else
 #endif
 		{
 			value = (char*)option->parameter;
 		}
+	}
 
-		if(type == 'F') {
-			/* call option parameter handler */
-			( ( void(*)(options_t *, char*, unsigned) )o->ptr )(opts, value, o->param);
-		} else {
-			/* save the option parameter */
-			*(char**)((char*)opts + ((char*)o->ptr - (char*)&opt)) = value;
-		}
-	} else if(o->type == 'f') {
+	/* process option, choosing the method by type */
+	switch(option_type) {
+	case F_UFLG:
+	case F_UENC:
+		*(unsigned*)((char*)opts + ((char*)o->ptr - (char*)&opt)) |= o->param;
+		break;
+	case F_CSTR:
+	case F_OPTH:
+		/* save the option parameter */
+		*(char**)((char*)opts + ((char*)o->ptr - (char*)&opt)) = value;
+		break;
+	case F_PFNC:
+		/* call option parameter handler */
+		( ( void(*)(options_t *, char*, unsigned) )o->ptr )(opts, value, o->param);
+		break;
+	case F_VFNC:
 		( ( void(*)(options_t *) )o->ptr )(opts); /* call option handler */
-	} else if(o->type == 'p') {
+		break;
+	case F_PRNT:
 		log_msg("%s", (char*)o->ptr);
 		rsh_exit(0);
+		break;
+	default:
+		assert(0); /* impossible option type */
 	}
 }
 
 /**
  * Search for config file.
- * 
+ *
  * @return the relative path to config file
  */
 static const char* find_conf_file(void)
@@ -501,7 +523,7 @@ static int read_config(void)
 		/* search for '=' */
 		index = strcspn(line, "=");
 		if(line[index] == 0) {
-			log_msg("%s: warning: can't parse line \"%s\"\n", conf_opt.config_file, line);
+			log_warning(_("%s: can't parse line \"%s\"\n"), conf_opt.config_file, line);
 			continue;
 		}
 		line[index] = 0;
@@ -514,7 +536,7 @@ static int read_config(void)
 		}
 
 		if(!t->type) {
-			log_msg("%s: warning: unknown option \"%s\"\n", conf_opt.config_file, line);
+			log_warning(_("%s: unknown option \"%s\"\n"), conf_opt.config_file, line);
 			continue;
 		}
 
@@ -524,6 +546,7 @@ static int read_config(void)
 		if(is_param_required(t->type)) {
 			rsh_vector_add_ptr(opt.mem, (value = rsh_strdup(value)));;
 		} else {
+			/* possible boolean values for a config file variable */
 			static const char* strings[] = {"1", "on", "yes", 0};
 			const char** cmp;
 			for(cmp = strings; *cmp && strcmp(value, *cmp); cmp++);
@@ -539,35 +562,26 @@ static int read_config(void)
 
 #ifdef _WIN32
 	if( (opt.flags & OPT_ENCODING) == 0 ) opt.flags |= (conf_opt.flags & OPT_ENCODING);
-	if( (opt.flags & OPT_ENCODING) == 0 ) opt.flags |= OPT_UTF8;
 #endif
 	return (res == 0 ? 0 : -1);
 }
 
-#ifdef _WIN32
-typedef wchar_t rhash_tchar;
-#define RSH_T(str) L##str
-#else
-typedef char rhash_tchar;
-#define RSH_T(str) str
-#endif
-
 /**
- * Find long option info, by it's name and retrive its parameter if required.
- * Error is reported for unknow options.
+ * Find long option info, by it's name and retrieve its parameter if required.
+ * Error is reported for unknown options.
  *
- * @param option structure to recieve the parsed option info
+ * @param option structure to receive the parsed option info
  * @param parg pointer to a command line argument
  */
-static void parse_long_option(parsed_option_t* option, rhash_tchar ***parg)
+static void parse_long_option(parsed_option_t* option, rsh_tchar ***parg)
 {
 	size_t length;
-	rhash_tchar* eq_sign;
+	rsh_tchar* eq_sign;
 	cmdline_opt_t *t;
 	char* name;
 
 #ifdef _WIN32
-	rhash_tchar* wname = **parg; /* skip "--" */
+	rsh_tchar* wname = **parg; /* "--<option name>" */
 	int fail = 0;
 	assert((**parg)[0] == L'-' && (**parg)[1] == L'-');
 
@@ -612,14 +626,18 @@ static void parse_long_option(parsed_option_t* option, rhash_tchar ***parg)
 	}
 }
 
+/**
+ * Parsed program command line.
+ */
 struct parsed_cmd_line_t
 {
 	blocks_vector_t options; /* array of parsed options */
+	int  argc;
 	char **argv;
-	int n_files;
-	rhash_tchar** files;
+	int n_files; /* the number of files */
+	rsh_tchar** files; /* array of files specified by command line */
 #ifdef _WIN32
-	rhash_tchar** warg;
+	rsh_tchar** warg; /* program arguments in Unicode */
 #endif
 };
 
@@ -632,32 +650,34 @@ static void parse_cmdline_options(struct parsed_cmd_line_t* cmd_line)
 {
 	int argc;
 	int n_files = 0, b_opt_end = 0;
-	rhash_tchar** files;
-	rhash_tchar **parg;
+	rsh_tchar** files;
+	rsh_tchar **parg, **end_arg;
 	parsed_option_t *next_opt;
 
 #ifdef _WIN32
 	parg = cmd_line->warg = CommandLineToArgvW(GetCommandLineW(), &argc);
-	if( NULL == parg ) {
-		die("CommandLineToArgvW failed\n");
+	if( NULL == parg || argc < 1) {
+		die(_("CommandLineToArgvW failed\n"));
 	}
 #else
+	argc = cmd_line->argc;
 	parg = cmd_line->argv;
-	for(argc = 0; parg[argc]; argc++); /* compute argc for files[] size */
 #endif
 
 	/* allocate array for files */
-	files = (rhash_tchar**)rsh_malloc(argc * sizeof(rhash_tchar*));
+	files = (rsh_tchar**)rsh_malloc(argc * sizeof(rsh_tchar*));
+	end_arg = parg + argc;
 
 	/* loop by program arguments */
-	for(parg++; *parg != NULL; parg++)
+	for(parg++; parg < end_arg; parg++)
 	{
 		/* if argument is not an option */
 		if((*parg)[0] != RSH_T('-') || (*parg)[1] == 0 || b_opt_end) {
-			/* file encountered, note that '-' is interpreted as stdin */
+			/* it's a file, note that '-' is interpreted as stdin */
 			files[n_files++] = *parg;
 			continue;
 		}
+
 		assert((*parg)[0] == RSH_T('-') && (*parg)[1] != 0);
 		
 		if((*parg)[1] == L'-' && (*parg)[2] == 0) {
@@ -678,16 +698,12 @@ static void parse_cmdline_options(struct parsed_cmd_line_t* cmd_line)
 			t = next_opt->o;
 
 			/* process encoding and -o/-l options early */
-			if(
-#ifdef _WIN32
-			(t->type == 'u' && t->ptr == &opt.flags && (t->param & OPT_ENCODING) != 0) ||
-#endif
-				(t->type == 'P' && (t->ptr == &opt.output || t->ptr == &opt.log))) {
+			if(is_output_modifier(t->type)) {
 				apply_option(&opt, next_opt);
 			}
 		} else if((*parg)[1] != 0) {
 			/* found '-'<some string> */
-			rhash_tchar* ptr;
+			rsh_tchar* ptr;
 			
 			/* parse short options. A string of several characters is interpreted
 			 * as separate short options */
@@ -717,17 +733,13 @@ static void parse_cmdline_options(struct parsed_cmd_line_t* cmd_line)
 					next_opt->parameter = (ptr[1] ? ptr + 1 : *(++parg));
 					if(!next_opt->parameter) {
 						/* note: need to check for parameter here, for early -o/-l options processing */
-						log_msg(PROGRAM_NAME ": argument is required for option %s\n", next_opt->name);
+						log_error(_("argument is required for option %s\n"), next_opt->name);
 						rsh_exit(2);
 					}
 				}
 
 				/* process encoding and -o/-l options early */
-				if(
-#ifdef _WIN32
-				(t->type == 'u' && t->ptr == &opt.flags && (t->param & OPT_ENCODING) != 0) ||
-#endif
-					(t->type == 'P' && (t->ptr == &opt.output || t->ptr == &opt.log))) {
+				if(is_output_modifier(t->type)) {
 					apply_option(&opt, next_opt);
 				}
 				if(next_opt->parameter) break;  /* a parameter ends the short options string */
@@ -754,16 +766,19 @@ static void apply_cmdline_options(struct parsed_cmd_line_t *cmd_line)
 		parsed_option_t* o = (parsed_option_t*)rsh_blocks_vector_get_ptr(
 			&cmd_line->options, i, 16, sizeof(parsed_option_t));
 
-		apply_option(&opt, o); /* process the option */
+		/* process the option, if it was not applied early */
+		if(!is_output_modifier(o->o->type)) {
+			apply_option(&opt, o);
+		}
 	}
 
 	/* copy formating options from config if not specified at command line */
-	if(!opt.printf && !opt.template_file && !opt.sum_flags && !opt.fmt) {
-		opt.printf = conf_opt.printf;
+	if(!opt.printf_str && !opt.template_file && !opt.sum_flags && !opt.fmt) {
+		opt.printf_str = conf_opt.printf_str;
 		opt.template_file = conf_opt.template_file;
 	}
 
-	if(!opt.printf && !opt.template_file) {
+	if(!opt.printf_str && !opt.template_file) {
 		if(!opt.fmt) opt.fmt = conf_opt.fmt;
 		if(!opt.sum_flags) opt.sum_flags = conf_opt.sum_flags;
 	}
@@ -862,17 +877,17 @@ static void make_final_options_checks(void)
 
 	if((opt.flags & OPT_VERBOSE) && conf_opt.config_file) {
 		/* note that the first log_msg call shall be made after setup_output() */
-		log_msg("Config file: %s\n", (conf_opt.config_file ? conf_opt.config_file : "None"));
+		log_msg(_("Config file: %s\n"), (conf_opt.config_file ? conf_opt.config_file : _("None")));
 	}
 
 	/* check that no more than one program mode specified */
 	if(opt.mode & (opt.mode - 1)) {
-		die(PROGRAM_NAME ": incompatible program modes\n");
+		die(_("incompatible program modes\n"));
 	}
 
-	ff = (opt.printf ? 1 : 0) | (opt.template_file ? 2 : 0) | (opt.fmt ? 4 : 0);
+	ff = (opt.printf_str ? 1 : 0) | (opt.template_file ? 2 : 0) | (opt.fmt ? 4 : 0);
 	if((opt.fmt & (opt.fmt - 1)) || (ff & (ff - 1))) {
-		die(PROGRAM_NAME ": too many formating options\n");
+		die(_("too many formating options\n"));
 	}
 
 	if(opt.mode & MODE_TORRENT) opt.sum_flags |= RHASH_BTIH;
@@ -887,7 +902,7 @@ static void make_final_options_checks(void)
  *
  * @param argv program arguments
  */
-void read_options(char *argv[])
+void read_options(int argc, char *argv[])
 {
 	struct parsed_cmd_line_t cmd_line;
 #ifdef _WIN32
@@ -903,16 +918,23 @@ void read_options(char *argv[])
 	memset(&cmd_line, 0, sizeof(cmd_line));
 	rsh_blocks_vector_init(&cmd_line.options);
 	cmd_line.argv = argv;
+	cmd_line.argc = argc;
 
 	/* parse command line and apply encoding options */
 	parse_cmdline_options(&cmd_line);
 	read_config();
+	
+#ifdef _WIN32
+	/* set default encoding if no encoding options were specified, */
+	/* this should be done here, even if config file was not found. */
+	if( (opt.flags & OPT_ENCODING) == 0 ) opt.flags |= OPT_UTF8;
+#endif
 
 	/* note: encoding and -o/-l options are already applied */
 	IF_WINDOWS(setup_console());
 	setup_output(); /* setup program output */
 
-	apply_cmdline_options(&cmd_line);
+	apply_cmdline_options(&cmd_line); /* process the rest of command options */
 
 	/* options were processed, so we don't need them anymore */
 	rsh_blocks_vector_destroy(&cmd_line.options);
@@ -925,19 +947,19 @@ void read_options(char *argv[])
 		wchar_t* path = cmd_line.files[i];
 		wchar_t* p = wcschr(path, L'\0') - 1;
 
-		/* strip trailing '\','/' symbols (if not preceeded by ':') */
+		/* strip trailing '\','/' symbols (if not preceded by ':') */
 		for(; p > path && IS_PATH_SEPARATOR_W(*p) && p[-1] != L':'; p--) *p = 0;
 		expand_wildcards(expanded_cnames, path);
 	}
-	/* the following NULL marks the end of files */
-	if(cmd_line.n_files) rsh_vector_add_ptr(expanded_cnames, NULL);
 
 	opt.cmd_vec = expanded_cnames;
 	opt.files = (char**)expanded_cnames->array;
+	opt.n_files = (int)expanded_cnames->size;
 	free(cmd_line.files);
 	LocalFree(cmd_line.warg);
 #else
 	opt.files = cmd_line.files;
+	opt.n_files = cmd_line.n_files;
 	rsh_vector_add_ptr(opt.mem, opt.files);
 #endif
 
